"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const DesignUsage_1 = require("./DesignUsage");
const QRData_1 = require("./QRData");
const ByteUtils_1 = require("./support/ByteUtils");
const ColorPalette_1 = require("./ColorPalette");
const IndexedImage_1 = require("./images/IndexedImage");
const CompositeIndexedImage_1 = require("./images/CompositeIndexedImage");
const CompositeImageLayout_1 = require("./CompositeImageLayout");
const ImageSize_1 = require("./ImageSize");
const QR_BYTE_LENGTH_NORMAL = 620;
const QR_BYTE_LENGTH_SPLIT = 540;
class Design {
    constructor(usage = DesignUsage_1.DesignUsage.CustomDesign) {
        /** The design's title (UTF-16 encoded, up to 42 bytes). */
        this.title = undefined;
        /** The player ID (2 bytes). */
        this.playerId = 5611;
        /** The creator's name (UTF-16 encoded, up to 18 bytes). */
        this.creator = undefined;
        /** The creator's sex (1 byte). */
        this.sex = 0;
        /** The village ID (2 bytes). */
        this.villageId = 59848;
        /** The village name (UTF-16 encoded, up to 18 bytes). */
        this.village = undefined;
        /** The language ID (1 byte). */
        this.languageId = 1;
        /** The country ID (1 byte). */
        this.countryId = 49;
        /** The region ID (1 byte). */
        this.regionId = 11;
        /** The color palette (15 bytes). */
        this.colorPalette = new ColorPalette_1.ColorPalette;
        /** The color (this field's purpose is currently unknown) (1 byte). */
        this.color = 0;
        /** The looks (this field's purpose is currently unknown) (1 byte). */
        this.looks = 0;
        /** The usage. */
        this._usage = DesignUsage_1.DesignUsage.CustomDesign;
        this._usage = usage;
        this.refreshImageData();
    }
    static fromBytes(bytes) {
        const design = new Design;
        design.loadBytes(bytes);
        return design;
    }
    static fromQRData(values) {
        const design = new Design;
        design.loadQRData(values);
        return design;
    }
    /** The colors used in the palette for this design */
    get colors() { return this.colorPalette.colors; }
    get usage() { return this._usage; }
    set usage(value) {
        this._usage = value;
        this.refreshImageData();
    }
    /** The usage ID. */
    get usageId() { return this.usage.id; }
    set usageId(value) {
        const usage = DesignUsage_1.DesignUsage.fromId(value);
        if (!usage) {
            throw Error('Invalid usage ID');
        }
        this.usage = usage;
    }
    get imageData() { return this._imageData; }
    refreshImageData() {
        const oldImageData = this._imageData;
        this._imageData = new IndexedImage_1.IndexedImage(this.usage.width, this.usage.height, this);
        if (oldImageData) {
            const newIndexes = this._imageData.colorIndexes;
            const oldIndexes = oldImageData.colorIndexes;
            newIndexes.set(oldIndexes.subarray(0, newIndexes.length));
        }
    }
    // Export
    toBytes() {
        if (this.title === undefined) {
            throw new Error('Title is required');
        }
        if (this.creator === undefined) {
            throw new Error('Creator name is required');
        }
        if (this.village === undefined) {
            throw new Error('Village name is required');
        }
        const data = new Uint8Array(this.usage.byteLength);
        ByteUtils_1.setString(data, 0, 42, this.title);
        ByteUtils_1.setNumber(data, 42, 2, this.playerId);
        ByteUtils_1.setString(data, 44, 18, this.creator);
        ByteUtils_1.setNumber(data, 62, 1, this.sex);
        ByteUtils_1.setNumber(data, 64, 2, this.villageId);
        ByteUtils_1.setString(data, 66, 18, this.village);
        ByteUtils_1.setNumber(data, 84, 1, this.languageId);
        ByteUtils_1.setNumber(data, 86, 1, this.countryId);
        ByteUtils_1.setNumber(data, 87, 1, this.regionId);
        ByteUtils_1.setByteSubset(data, 88, this.colorPalette.acnlBytes);
        ByteUtils_1.setNumber(data, 103, 1, this.color);
        ByteUtils_1.setNumber(data, 104, 1, this.looks);
        ByteUtils_1.setNumber(data, 105, 1, this.usageId);
        ByteUtils_1.setByteSubset(data, 108, ByteUtils_1.compressImageData(this.imageData.colorIndexes));
        return data;
    }
    toQRData() {
        const bytes = this.toBytes();
        if (this.usage.qrCodeCount === 4) {
            const parity = ByteUtils_1.calculateParity(bytes);
            return [
                new QRData_1.QRData(ByteUtils_1.getByteSubset(bytes, 0, 540), 0, parity),
                new QRData_1.QRData(ByteUtils_1.getByteSubset(bytes, 540, 540), 1, parity),
                new QRData_1.QRData(ByteUtils_1.getByteSubset(bytes, 1080, 540), 2, parity),
                new QRData_1.QRData(ByteUtils_1.getByteSubset(bytes, 1620, 540), 3, parity),
            ];
        }
        return [
            new QRData_1.QRData(bytes),
        ];
    }
    // Import
    loadBytes(bytes) {
        this.loadMetadataBytes(bytes);
        this.imageData.colorIndexes = ByteUtils_1.extractImageData(ByteUtils_1.getByteSubset(bytes, 108, this.usage.imageDataByteLength));
    }
    loadMetadataBytes(data) {
        this.title = ByteUtils_1.getString(data, 0, 42);
        this.playerId = ByteUtils_1.getNumber(data, 42, 2);
        this.creator = ByteUtils_1.getString(data, 44, 18);
        this.sex = ByteUtils_1.getNumber(data, 62, 1);
        this.villageId = ByteUtils_1.getNumber(data, 64, 2);
        this.village = ByteUtils_1.getString(data, 66, 18);
        this.languageId = ByteUtils_1.getNumber(data, 84, 1);
        this.countryId = ByteUtils_1.getNumber(data, 86, 1);
        this.regionId = ByteUtils_1.getNumber(data, 87, 1);
        this.colorPalette.acnlBytes = ByteUtils_1.getByteSubset(data, 88, 15);
        this.color = ByteUtils_1.getNumber(data, 103, 1);
        this.looks = ByteUtils_1.getNumber(data, 104, 1);
        this.usageId = ByteUtils_1.getNumber(data, 105, 1);
    }
    loadQRData(values) {
        if (values instanceof QRData_1.QRData) {
            values = [values];
        }
        // For normal QR data, we can just pass along the single QR code's data
        const normalData = values.find(d => d.bytes.length === QR_BYTE_LENGTH_NORMAL);
        if (normalData) {
            return this.loadBytes(normalData.bytes);
        }
        // For pro data, we need to apply metadata from the first code and color data from all four
        values.filter(d => d.bytes.length === QR_BYTE_LENGTH_SPLIT)
            .forEach(d => this.loadProQRData(d));
    }
    loadProQRData(data) {
        switch (data.index) {
            case 0:
                this.loadMetadataBytes(data.bytes);
                ByteUtils_1.setByteSubset(this.imageData.colorIndexes, 0, ByteUtils_1.extractImageData(ByteUtils_1.getByteSubset(data.bytes, 108, 432)));
                break;
            case 1:
                ByteUtils_1.setByteSubset(this.imageData.colorIndexes, 864, ByteUtils_1.extractImageData(ByteUtils_1.getByteSubset(data.bytes, 0, 540)));
                break;
            case 2:
                ByteUtils_1.setByteSubset(this.imageData.colorIndexes, 1944, ByteUtils_1.extractImageData(ByteUtils_1.getByteSubset(data.bytes, 0, 540)));
                break;
            case 3:
                ByteUtils_1.setByteSubset(this.imageData.colorIndexes, 3024, ByteUtils_1.extractImageData(ByteUtils_1.getByteSubset(data.bytes, 0, 536)));
                break;
        }
    }
    // Image conversion
    getIndexedImage(layout = CompositeImageLayout_1.CompositeImageLayout.Normal) {
        // For pro designs, lay out the image segments in a 64x64 image
        if (this.usage.imageSize === ImageSize_1.ImageSize.Large) {
            // ACPatterns.com uses a slightly different image layout than the game itself
            if (layout === CompositeImageLayout_1.CompositeImageLayout.ACPatterns) {
                return (new CompositeIndexedImage_1.CompositeIndexedImage(64, 64))
                    .addSource(this.imageData.getSegment(0, 0, 32, 32), 0, 0)
                    .addSource(this.imageData.getSegment(0, 32, 32, 32), 0, 32)
                    .addSource(this.imageData.getSegment(0, 64, 32, 32), 32, 0)
                    .addSource(this.imageData.getSegment(0, 96, 32, 32), 32, 32);
            }
            return (new CompositeIndexedImage_1.CompositeIndexedImage(64, 64))
                .addSource(this.imageData.getSegment(0, 0, 32, 32), 32, 0)
                .addSource(this.imageData.getSegment(0, 32, 32, 32), 0, 0)
                .addSource(this.imageData.getSegment(0, 64, 32, 16), 0, 48)
                .addSource(this.imageData.getSegment(0, 80, 32, 16), 32, 48)
                .addSource(this.imageData.getSegment(0, 96, 32, 16), 32, 32)
                .addSource(this.imageData.getSegment(0, 112, 32, 16), 0, 32);
        }
        // For normal designs, we can just return the original image
        return this.imageData;
    }
    toImage(layout = CompositeImageLayout_1.CompositeImageLayout.Normal) {
        return this.getIndexedImage(layout).toImage();
    }
}
exports.Design = Design;
